import email
import hashlib
import json
import logging
import os
import requests
import smtplib
import ssl

import yaml

from mastodon import Mastodon

# On paramètre le niveau de verbosité des logs
logging.basicConfig(level=os.environ.get('LOG_LEVEL', 'WARNING').upper())
logger = logging.getLogger(__name__)


class Attrap_bot:
    class SMTP_sender:
        port = 587
        ssl = False
        starttls = True

        def __init__(self, hostname):
            self.hostname = hostname

        def set_username(self, username):
            self.username = username

        def set_password(self, password):
            self.password = password

        def set_port(self, port):
            if not isinstance(port, int):
                self.port = int(port)
            else:
                self.port = port

        def set_ssl(self, ssl):
            self.ssl = ssl
            if ssl is True:
                self.starttls = False

        def set_starttls(self, starttls):
            self.starttls = starttls
            if starttls is True:
                self.ssl = False

        def send(self, email_from, email_to, email_subject, email_message):
            message = email.message.EmailMessage()

            message['Subject'] = email_subject
            message['From'] = email_from
            message['Date'] = email.utils.formatdate()
            message.set_content(email_message)

            if self.ssl:
                smtp = smtplib.SMTP_SSL(self.hostname, self.port, context=ssl.create_default_context())
            elif self.starttls:
                smtp = smtplib.SMTP(self.hostname, self.port)
                smtp.starttls(context=ssl.create_default_context())
            else:
                smtp = smtplib.SMTP(self.hostname, self.port)

            if self.username:
                smtp.login(self.username, self.password)

            for address in email_to.split(','):
                del message['To']
                message['To'] = address
                del message['Message-ID']
                message['Message-ID'] = email.utils.make_msgid(domain=email_from.split('@')[-1])
                smtp.send_message(message)

            smtp.quit()

    class Mastodon_sender:
        mastodon = None

        def __init__(self, instance, access_token):
            self.instance = instance
            self.access_token = access_token

        def send(self, message):
            if not self.mastodon:
                self.mastodon = Mastodon(
                    api_base_url=self.instance,
                    access_token=self.access_token
                )

            self.mastodon.toot(message)

    def __init__(self, config):
        # On surcharge la configuration par défaut avec la configuration donnée
        config_default = yaml.load(open('config.default.yml', 'r'), Loader=yaml.FullLoader)
        self.config = config_default | config

        logger.info('Démarrage de Attrap_bot')
        logger.info(f"Source des données : {self.config['data_source']}")

    def analyze(self):
        if self.config.get('queries'):
            for i in self.config['queries']:
                query = self.config['queries'][i]
                query_id = i
                search = query['search']
                administration = ""
                if query.get('administration') and query['administration']:
                    administration = query['administration']

                logger.info(f'Démarrage de la recherche {query_id}')
                logger.info(f'Requête : {search}')
                logger.info(f'Administration : {administration}')

                # On récupère de quoi envoyer par mail et sur Mastodon
                smtp_sender = None
                mastodon_sender = None
                if query.get('email') and query['email'].get('smtp'):
                    logger.debug('Configuration SMTP trouvée')
                    smtp_sender = self.get_smtp_sender(query['email']['smtp'])
                if query.get('mastodon') and query['mastodon'].get('instance'):
                    logger.debug('Configuration Mastodon trouvée')
                    mastodon_sender = self.get_mastodon_sender(query['mastodon']['instance'])

                # On fabrique l'URL de requête
                request_url = self.config['data_source'].replace('{search}', search).replace('{administration}', administration)

                # On ouvre le fichier de status
                status_file_path = f'{query_id}.latest.txt'
                if os.path.isfile(status_file_path):
                    status = open(status_file_path, 'r').read().strip()
                else:
                    status = None

                # On interroge l'API d'Attrap
                response = json.loads(requests.get(request_url, timeout=(10, 120)).content)
                raa = response['elements']

                # Si le dernier RAA n'est pas celui connu, on analyse les résultats
                latest = raa[0]
                if latest['id'] != status:
                    email_message = ""

                    # On récupère les derniers RAA jusqu'à ce qu'on trouve le dernier analysé
                    raa_candidates = []
                    for result in raa:
                        if result['id'] != status:
                            raa_candidates.append(result)
                        else:
                            break

                    # Maintenant, on prend les derniers RAA à analyser, pour commencer par les plus anciens en premier
                    for result in raa_candidates[::-1]:
                        status = result['id']
                        raa_name = result['name']
                        raa_date = result['date']
                        raa_administration = result['administration']
                        raa_administration_name = result['administration_name']
                        raa_url = result['url']

                        # On affiche le résultat dans la console
                        if (self.config['console_output']):
                            print(f'\033[92m{raa_name}\033[0m ({raa_date}) : {raa_url}')

                        # On publie sur Mastodon
                        if mastodon_sender:
                            toot = f'{raa_name} ({raa_date})\n\n{raa_url}'
                            if query['mastodon']['prefix'] and not query['mastodon']['prefix'] == '':
                                prefix = query['mastodon']['prefix']
                                prefix = prefix.replace('{id}', query_id)
                                prefix = prefix.replace('{administration}', raa_administration)
                                prefix = prefix.replace('{administration_name}', raa_administration_name)
                                toot = f"{prefix}\n\n{toot}"
                            if query['mastodon']['suffix'] and not query['mastodon']['suffix'] == '':
                                suffix = query['mastodon']['suffix']
                                suffix = suffix.replace('{id}', query_id)
                                suffix = suffix.replace('{administration}', raa_administration)
                                suffix = suffix.replace('{administration_name}', raa_administration_name)
                                toot = f"{toot}\n\n{suffix}"
                            mastodon_sender.send(toot)

                        # On complète le mail
                        if smtp_sender:
                            if email_message == "":
                                email_message = f"Attrap ({query_id})\nRequête : {search}"
                            email_message = f"{email_message}\n\n{raa_administration_name} :\n{raa_name} ({raa_date})\nURL : {raa_url}"

                    # On envoie le mail si le message est prêt
                    if not email_message == "":
                        smtp_sender.send(
                            query['email']['from'],
                            query['email']['to'],
                            f"[Attrap] [{query_id}] Nouveaux éléments trouvés",
                            email_message
                        )

                    if os.path.isfile(status_file_path):
                        os.remove(status_file_path)
                    status_file = open(status_file_path, 'w')
                    status_file.write(status)
                    status_file.close()

    def get_mastodon_sender(self, config_id):
        if self.config.get('mastodon'):
            if self.config['mastodon'].get(config_id) and self.config['mastodon'][config_id]['instance'] and self.config['mastodon'][config_id]['access_token']:
                return Attrap_bot.Mastodon_sender(self.config['mastodon'][config_id]['instance'], self.config['mastodon'][config_id]['access_token'])
        logger.warning(f"La configuration Mastodon est invalide (id: {config_id})")
        return None

    def get_smtp_sender(self, config_id):
        if self.config.get('smtp'):
            if self.config['smtp'].get(config_id) and self.config['smtp'][config_id]['hostname']:
                smtp_sender = Attrap_bot.SMTP_sender(self.config['smtp'][config_id]['hostname'])
                if self.config['smtp'][config_id].get('username'):
                    smtp_sender.set_username(self.config['smtp'][config_id]['username'])
                if self.config['smtp'][config_id].get('password'):
                    smtp_sender.set_password(self.config['smtp'][config_id]['password'])
                if self.config['smtp'][config_id].get('port'):
                    smtp_sender.set_port(self.config['smtp'][config_id]['port'])
                if self.config['smtp'][config_id].get('ssl'):
                    smtp_sender.set_ssl(self.config['smtp'][config_id]['ssl'])
                if self.config['smtp'][config_id].get('starttls'):
                    smtp_sender.set_starttls(self.config['smtp'][config_id]['starttls'])
                return smtp_sender
        logger.warning(f"La configuration Mastodon est invalide (id: {config_id})")
        return None
